@@ -1,15 +1,12 @@
 /*
- *  Copyright (c) 2017 Dmitry Kazakov <dimula73@gmail.com>
+ *  SPDX-FileCopyrightText: 2017 Dmitry Kazakov <dimula73@gmail.com>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+v *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
@@ -22,15 +19,14 @@
 #include "KoSvgText.h"
 #include "KoSvgTextProperties.h"
 
-
 #include "kis_debug.h"
-#include <KoXmlReader.h>
 #include <KoXmlWriter.h>
 #include <KoXmlNS.h>
 
 #include <SvgLoadingContext.h>
 #include <SvgGraphicContext.h>
 #include <SvgUtil.h>
+#include <SimpleShapeContainerModel.h>
 
 #include <SvgSavingContext.h>
 #include <SvgStyleWriter.h>
@@ -41,6 +37,7 @@
 
 #include <html/HtmlSavingContext.h>
 
+#include <FlakeDebug.h>
 
 namespace {
 
@@ -117,25 +114,73 @@ QString convertListAttribute(const QVector<qreal> &values) {
 }
 }
 
-struct KoSvgTextChunkShapePrivate::LayoutInterface : public KoSvgTextChunkShapeLayoutInterface
+#include <ksharedconfig.h>
+#include <kconfiggroup.h>
+
+/**
+ * HACK ALERT: this is a function from a private Qt's header qfont_p.h,
+ * we don't include the whole header, because it is painful in the
+ * environments we don't fully control, e.g. in distribution packages.
+ */
+Q_GUI_EXPORT int qt_defaultDpi();
+
+namespace {
+int forcedDpiForQtFontBugWorkaround() {
+    KConfigGroup cfg(KSharedConfig::openConfig(), "");
+    int value = cfg.readEntry("forcedDpiForQtFontBugWorkaround", qt_defaultDpi());
+
+    if (value < 0) {
+        value = qt_defaultDpi();
+    }
+
+    return value;
+}
+
+
+KoSvgTextProperties adjustPropertiesForFontSizeWorkaround(const KoSvgTextProperties &properties)
+{
+    if (!properties.hasProperty(KoSvgTextProperties::FontSizeId)) return properties;
+
+    KoSvgTextProperties result = properties;
+
+    const int forcedFontDPI = forcedDpiForQtFontBugWorkaround();
+
+    if (result.hasProperty(KoSvgTextProperties::KraTextVersionId) &&
+        result.property(KoSvgTextProperties::KraTextVersionId).toInt() < 2 &&
+        forcedFontDPI > 0) {
+
+        qreal fontSize = result.property(KoSvgTextProperties::FontSizeId).toReal();
+        fontSize *= qreal(forcedFontDPI) / 72.0;
+        result.setProperty(KoSvgTextProperties::FontSizeId, fontSize);
+    }
+
+    result.setProperty(KoSvgTextProperties::KraTextVersionId, 2);
+
+    return result;
+}
+
+}
+
+
+struct KoSvgTextChunkShape::Private::LayoutInterface : public KoSvgTextChunkShapeLayoutInterface
 {
     LayoutInterface(KoSvgTextChunkShape *_q) : q(_q) {}
 
-    KoSvgText::AutoValue textLength() const {
-        return q->d_func()->textLength;
+    KoSvgText::AutoValue textLength() const override {
+        return q->s->textLength;
     }
 
-    KoSvgText::LengthAdjust lengthAdjust() const {
-        return q->d_func()->lengthAdjust;
+    KoSvgText::LengthAdjust lengthAdjust() const override {
+        return q->s->lengthAdjust;
     }
 
-    int numChars() const {
-        KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(!q->shapeCount() || q->d_func()->text.isEmpty(), 0);
+    int numChars() const override {
+        KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(!q->shapeCount() || q->s->text.isEmpty(), 0);
 
         int result = 0;
 
         if (!q->shapeCount()) {
-            result = q->d_func()->text.size();
+            result = q->s->text.size();
         } else {
             Q_FOREACH (KoShape *shape, q->shapes()) {
                 KoSvgTextChunkShape *chunkShape = dynamic_cast<KoSvgTextChunkShape*>(shape);
@@ -147,7 +192,7 @@ struct KoSvgTextChunkShapePrivate::LayoutInterface : public KoSvgTextChunkShapeL
         return result;
     }
 
-    int relativeCharPos(KoSvgTextChunkShape *child, int pos) const {
+    int relativeCharPos(KoSvgTextChunkShape *child, int pos) const override {
         QList<KoShape*> childShapes = q->shapes();
 
         int result = -1;
@@ -169,21 +214,21 @@ struct KoSvgTextChunkShapePrivate::LayoutInterface : public KoSvgTextChunkShapeL
         return result;
     }
 
-    bool isTextNode() const {
-        KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(!q->shapeCount() || q->d_func()->text.isEmpty(), false);
+    bool isTextNode() const override {
+        KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(!q->shapeCount() || q->s->text.isEmpty(), false);
         return !q->shapeCount();
     }
 
-    QString nodeText() const {
-        KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(!q->shapeCount() || q->d_func()->text.isEmpty(), 0);
-        return !q->shapeCount() ? q->d_func()->text : QString();
+    QString nodeText() const override {
+        KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(!q->shapeCount() || q->s->text.isEmpty(), 0);
+        return !q->shapeCount() ? q->s->text : QString();
     }
 
-    QVector<KoSvgText::CharTransformation> localCharTransformations() const {
+    QVector<KoSvgText::CharTransformation> localCharTransformations() const override {
         KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(isTextNode(), QVector<KoSvgText::CharTransformation>());
 
-        const QVector<KoSvgText::CharTransformation> t = q->d_func()->localTransformations;
-        return t.mid(0, qMin(t.size(), q->d_func()->text.size()));
+        const QVector<KoSvgText::CharTransformation> t = q->s->localTransformations;
+        return t.mid(0, qMin(t.size(), q->s->text.size()));
     }
 
     static QString getBidiOpening(KoSvgText::Direction direction, KoSvgText::UnicodeBidi bidi) {
@@ -200,13 +245,13 @@ struct KoSvgTextChunkShapePrivate::LayoutInterface : public KoSvgTextChunkShapeL
         return result;
     }
 
-    QVector<SubChunk> collectSubChunks() const {
+    QVector<SubChunk> collectSubChunks() const override {
         QVector<SubChunk> result;
 
         if (isTextNode()) {
-            const QString text = q->d_func()->text;
-            const KoSvgText::KoSvgCharChunkFormat format = q->d_func()->fetchCharFormat();
-            QVector<KoSvgText::CharTransformation> transforms = q->d_func()->localTransformations;
+            const QString text = q->s->text;
+            const KoSvgText::KoSvgCharChunkFormat format = q->fetchCharFormat();
+            QVector<KoSvgText::CharTransformation> transforms = q->s->localTransformations;
 
             /**
              * Sometimes SVG can contain the X,Y offsets for the pieces of text that
@@ -216,8 +261,8 @@ struct KoSvgTextChunkShapePrivate::LayoutInterface : public KoSvgTextChunkShapeL
                 transforms.resize(text.size());
             }
 
-            KoSvgText::UnicodeBidi bidi = KoSvgText::UnicodeBidi(q->d_func()->properties.propertyOrDefault(KoSvgTextProperties::UnicodeBidiId).toInt());
-            KoSvgText::Direction direction = KoSvgText::Direction(q->d_func()->properties.propertyOrDefault(KoSvgTextProperties::DirectionId).toInt());
+            KoSvgText::UnicodeBidi bidi = KoSvgText::UnicodeBidi(q->s->properties.propertyOrDefault(KoSvgTextProperties::UnicodeBidiId).toInt());
+            KoSvgText::Direction direction = KoSvgText::Direction(q->s->properties.propertyOrDefault(KoSvgTextProperties::DirectionId).toInt());
             const QString bidiOpening = getBidiOpening(direction, bidi);
 
             if (!bidiOpening.isEmpty()) {
@@ -269,23 +314,23 @@ struct KoSvgTextChunkShapePrivate::LayoutInterface : public KoSvgTextChunkShapeL
         KIS_SAFE_ASSERT_RECOVER_RETURN(isTextNode());
         QPainterPath path;
         path.addRect(rect);
-        path |= q->d_func()->associatedOutline;
+        path |= q->s->associatedOutline;
         path.setFillRule(Qt::WindingFill);
         path = path.simplified();
 
-        q->d_func()->associatedOutline = path;
-        q->d_func()->size = path.boundingRect().size();
+        q->s->associatedOutline = path;
+        q->setSize(path.boundingRect().size());
 
         q->notifyChanged();
-        q->d_func()->shapeChanged(KoShape::SizeChanged);
+        q->shapeChangedPriv(KoShape::SizeChanged);
     }
 
     void clearAssociatedOutline() override {
-        q->d_func()->associatedOutline = QPainterPath();
-        q->d_func()->size = QSizeF();
+        q->s->associatedOutline = QPainterPath();
+        q->setSize(QSizeF());
 
         q->notifyChanged();
-        q->d_func()->shapeChanged(KoShape::SizeChanged);
+        q->shapeChangedPriv(KoShape::SizeChanged);
     }
 
 private:
@@ -293,24 +338,24 @@ private:
 };
 
 KoSvgTextChunkShape::KoSvgTextChunkShape()
-    : KoShapeContainer(new KoSvgTextChunkShapePrivate(this))
+    : KoShapeContainer()
+    , d(new Private)
+    , s(new SharedData)
 {
-    Q_D(KoSvgTextChunkShape);
-    d->layoutInterface.reset(new KoSvgTextChunkShapePrivate::LayoutInterface(this));
+    d->layoutInterface.reset(new KoSvgTextChunkShape::Private::LayoutInterface(this));
 }
 
 KoSvgTextChunkShape::KoSvgTextChunkShape(const KoSvgTextChunkShape &rhs)
-    : KoShapeContainer(new KoSvgTextChunkShapePrivate(*rhs.d_func(), this))
+    : KoShapeContainer(rhs)
+    , d(new Private)
+    , s(rhs.s)
 {
-    Q_D(KoSvgTextChunkShape);
-    d->layoutInterface.reset(new KoSvgTextChunkShapePrivate::LayoutInterface(this));
-}
-
-KoSvgTextChunkShape::KoSvgTextChunkShape(KoSvgTextChunkShapePrivate *dd)
-    : KoShapeContainer(dd)
-{
-    Q_D(KoSvgTextChunkShape);
-    d->layoutInterface.reset(new KoSvgTextChunkShapePrivate::LayoutInterface(this));
+    if (rhs.model()) {
+        SimpleShapeContainerModel *otherModel = dynamic_cast<SimpleShapeContainerModel*>(rhs.model());
+        KIS_ASSERT_RECOVER_RETURN(otherModel);
+        setModelInit(new SimpleShapeContainerModel(*otherModel));
+    }
+    d->layoutInterface.reset(new KoSvgTextChunkShape::Private::LayoutInterface(this));
 }
 
 KoSvgTextChunkShape::~KoSvgTextChunkShape()
@@ -340,13 +385,11 @@ QRectF KoSvgTextChunkShape::outlineRect() const
 
 QPainterPath KoSvgTextChunkShape::outline() const
 {
-    Q_D(const KoSvgTextChunkShape);
-
     QPainterPath result;
     result.setFillRule(Qt::WindingFill);
 
     if (d->layoutInterface->isTextNode()) {
-        result = d->associatedOutline;
+        result = s->associatedOutline;
     } else {
         Q_FOREACH (KoShape *shape, shapes()) {
             KoSvgTextChunkShape *chunkShape = dynamic_cast<KoSvgTextChunkShape*>(shape);
@@ -359,29 +402,13 @@ QPainterPath KoSvgTextChunkShape::outline() const
     return result.simplified();
 }
 
-void KoSvgTextChunkShape::paintComponent(QPainter &painter, const KoViewConverter &converter, KoShapePaintingContext &paintContext)
+void KoSvgTextChunkShape::paintComponent(QPainter &painter) const
 {
     Q_UNUSED(painter);
-    Q_UNUSED(converter);
-    Q_UNUSED(paintContext);
-}
-
-void KoSvgTextChunkShape::saveOdf(KoShapeSavingContext &context) const
-{
-    Q_UNUSED(context);
-}
-
-bool KoSvgTextChunkShape::loadOdf(const KoXmlElement &element, KoShapeLoadingContext &context)
-{
-    Q_UNUSED(element);
-    Q_UNUSED(context);
-    return false;
 }
 
 bool KoSvgTextChunkShape::saveHtml(HtmlSavingContext &context)
 {
-    Q_D(KoSvgTextChunkShape);
-
     // Should we add a newline? Check for vertical movement if we're using rtl or ltr text
     // XXX: if vertical text, check horizontal movement.
     QVector<qreal> xPos;
@@ -390,10 +417,10 @@ bool KoSvgTextChunkShape::saveHtml(HtmlSavingContext &context)
     QVector<qreal> dyPos;
     QVector<qreal> rotate;
 
-    fillTransforms(&xPos, &yPos, &dxPos, &dyPos, &rotate, d->localTransformations);
+    fillTransforms(&xPos, &yPos, &dxPos, &dyPos, &rotate, s->localTransformations);
 
-    for (int i = 0; i < d->localTransformations.size(); i++) {
-        const KoSvgText::CharTransformation &t = d->localTransformations[i];
+    for (int i = 0; i < s->localTransformations.size(); i++) {
+        const KoSvgText::CharTransformation &t = s->localTransformations[i];
 
         appendLazy(&xPos, t.xPos, i, false);
         appendLazy(&yPos, t.yPos, i, false);
@@ -415,7 +442,7 @@ bool KoSvgTextChunkShape::saveHtml(HtmlSavingContext &context)
         }
         // XXX: Save the style?
 
-    } else if (parent->isRootTextNode()) {
+    } else if (parent && parent->isRootTextNode()) {
         context.shapeWriter().startElement("p", false);
     } else {
         context.shapeWriter().startElement("span", false);
@@ -446,9 +473,6 @@ bool KoSvgTextChunkShape::saveHtml(HtmlSavingContext &context)
                         .append(";" );
             } else if (QString(it.key().toLatin1().data()).contains("font-size")){
                 QString val = it.value();
-                if (QRegExp ("\\d*").exactMatch(val)) {
-                    val.append("pt");
-                }
                 styleString.append(it.key().toLatin1().data())
                         .append(": ")
                         .append(val)
@@ -463,9 +487,9 @@ bool KoSvgTextChunkShape::saveHtml(HtmlSavingContext &context)
         context.shapeWriter().addAttribute("style", styleString);
     }
     if (layoutInterface()->isTextNode()) {
-        qDebug() << "saveHTML" << this << d->text << xPos << yPos << dxPos << dyPos;
+        debugFlake << "saveHTML" << this << s->text << xPos << yPos << dxPos << dyPos;
         // After adding all the styling to the <p> element, add the text
-        context.shapeWriter().addTextNode(d->text);
+        context.shapeWriter().addTextNode(s->text);
     }
     else {
         Q_FOREACH (KoShape *child, this->shapes()) {
@@ -494,13 +518,16 @@ void writeTextListAttribute(const QString &attribute, const QVector<qreal> &valu
 
 bool KoSvgTextChunkShape::saveSvg(SvgSavingContext &context)
 {
-    Q_D(KoSvgTextChunkShape);
-
     if (isRootTextNode()) {
         context.shapeWriter().startElement("text", false);
 
         if (!context.strippedTextMode()) {
             context.shapeWriter().addAttribute("id", context.getID(this));
+            context.shapeWriter().addAttribute("krita:useRichText", s->isRichTextPreferred ? "true" : "false");
+
+            // save the version to distinguish from the buggy Krita version
+            context.shapeWriter().addAttribute("krita:textVersion", 2);
+
             SvgUtil::writeTransformAttributeLazy("transform", transformation(), context.shapeWriter());
             SvgStyleWriter::saveSvgStyle(this, context);
         } else {
@@ -522,7 +549,7 @@ bool KoSvgTextChunkShape::saveSvg(SvgSavingContext &context)
         QVector<qreal> dyPos;
         QVector<qreal> rotate;
 
-        fillTransforms(&xPos, &yPos, &dxPos, &dyPos, &rotate, d->localTransformations);
+        fillTransforms(&xPos, &yPos, &dxPos, &dyPos, &rotate, s->localTransformations);
 
         writeTextListAttribute("x", xPos, context.shapeWriter());
         writeTextListAttribute("y", yPos, context.shapeWriter());
@@ -531,10 +558,10 @@ bool KoSvgTextChunkShape::saveSvg(SvgSavingContext &context)
         writeTextListAttribute("rotate", rotate, context.shapeWriter());
     }
 
-    if (!d->textLength.isAuto) {
-        context.shapeWriter().addAttribute("textLength", KisDomUtils::toString(d->textLength.customValue));
+    if (!s->textLength.isAuto) {
+        context.shapeWriter().addAttribute("textLength", KisDomUtils::toString(s->textLength.customValue));
 
-        if (d->lengthAdjust == KoSvgText::LengthAdjustSpacingAndGlyphs) {
+        if (s->lengthAdjust == KoSvgText::LengthAdjustSpacingAndGlyphs) {
             context.shapeWriter().addAttribute("lengthAdjust", "spacingAndGlyphs");
         }
     }
@@ -545,6 +572,8 @@ bool KoSvgTextChunkShape::saveSvg(SvgSavingContext &context)
 
     KoSvgTextProperties ownProperties = textProperties().ownProperties(parentProperties);
 
+    ownProperties = adjustPropertiesForFontSizeWorkaround(ownProperties);
+
     // we write down stroke/fill iff they are different from the parent's value
     if (!isRootTextNode()) {
         if (ownProperties.hasProperty(KoSvgTextProperties::FillId)) {
@@ -562,7 +591,7 @@ bool KoSvgTextChunkShape::saveSvg(SvgSavingContext &context)
     }
 
     if (layoutInterface()->isTextNode()) {
-        context.shapeWriter().addTextNode(d->text);
+        context.shapeWriter().addTextNode(s->text);
     } else {
         Q_FOREACH (KoShape *child, this->shapes()) {
             KoSvgTextChunkShape *childText = dynamic_cast<KoSvgTextChunkShape*>(child);
@@ -577,29 +606,25 @@ bool KoSvgTextChunkShape::saveSvg(SvgSavingContext &context)
     return true;
 }
 
+#include <SvgGraphicContext.h>
 
-void KoSvgTextChunkShapePrivate::loadContextBasedProperties(SvgGraphicsContext *gc)
+
+void KoSvgTextChunkShape::SharedData::loadContextBasedProperties(SvgGraphicsContext *gc)
 {
     properties = gc->textProperties;
-    font = gc->font;
-    fontFamiliesList = gc->fontFamiliesList;
 }
 
 void KoSvgTextChunkShape::resetTextShape()
 {
-    Q_D(KoSvgTextChunkShape);
-
     using namespace KoSvgText;
 
-    d->properties = KoSvgTextProperties();
-    d->font = QFont();
-    d->fontFamiliesList = QStringList();
+    s->properties = KoSvgTextProperties();
 
-    d->textLength = AutoValue();
-    d->lengthAdjust = LengthAdjustSpacing;
+    s->textLength = AutoValue();
+    s->lengthAdjust = LengthAdjustSpacing;
 
-    d->localTransformations.clear();
-    d->text.clear();
+    s->localTransformations.clear();
+    s->text.clear();
 
 
     // all the subchunks are destroyed!
@@ -611,17 +636,15 @@ void KoSvgTextChunkShape::resetTextShape()
     }
 }
 
-bool KoSvgTextChunkShape::loadSvg(const KoXmlElement &e, SvgLoadingContext &context)
+bool KoSvgTextChunkShape::loadSvg(const QDomElement &e, SvgLoadingContext &context)
 {
-    Q_D(KoSvgTextChunkShape);
-
     SvgGraphicsContext *gc = context.currentGC();
     KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(gc, false);
 
-    d->loadContextBasedProperties(gc);
+    s->loadContextBasedProperties(gc);
 
-    d->textLength = KoSvgText::parseAutoValueXY(e.attribute("textLength", ""), context, "");
-    d->lengthAdjust = KoSvgText::parseLengthAdjust(e.attribute("lengthAdjust", "spacing"));
+    s->textLength = KoSvgText::parseAutoValueXY(e.attribute("textLength", ""), context, "");
+    s->lengthAdjust = KoSvgText::parseLengthAdjust(e.attribute("lengthAdjust", "spacing"));
 
     QVector<qreal> xPos = parseListAttributeX(e.attribute("x", ""), context);
     QVector<qreal> yPos = parseListAttributeY(e.attribute("y", ""), context);
@@ -634,22 +657,22 @@ bool KoSvgTextChunkShape::loadSvg(const KoXmlElement &e, SvgLoadingContext &cont
                   dxPos.size(), dyPos.size(),
                   rotate.size()});
 
-    d->localTransformations.resize(numLocalTransformations);
+    s->localTransformations.resize(numLocalTransformations);
     for (int i = 0; i < numLocalTransformations; i++) {
         if (i < xPos.size()) {
-            d->localTransformations[i].xPos = xPos[i];
+            s->localTransformations[i].xPos = xPos[i];
         }
         if (i < yPos.size()) {
-            d->localTransformations[i].yPos = yPos[i];
+            s->localTransformations[i].yPos = yPos[i];
         }
         if (i < dxPos.size() && dxPos[i] != 0.0) {
-            d->localTransformations[i].dxPos = dxPos[i];
+            s->localTransformations[i].dxPos = dxPos[i];
         }
         if (i < dyPos.size() && dyPos[i] != 0.0) {
-            d->localTransformations[i].dyPos = dyPos[i];
+            s->localTransformations[i].dyPos = dyPos[i];
         }
         if (i < rotate.size()) {
-            d->localTransformations[i].rotate = rotate[i];
+            s->localTransformations[i].rotate = rotate[i];
         }
     }
 
@@ -659,7 +682,7 @@ bool KoSvgTextChunkShape::loadSvg(const KoXmlElement &e, SvgLoadingContext &cont
 namespace {
 
 QString cleanUpString(QString text) {
-    text.replace(QRegExp("[\\r\\n]"), "");
+    text.replace(QRegExp("[\\r\\n\u2028]"), "");
     text.replace(QRegExp(" {2,}"), " ");
     return text;
 }
@@ -670,11 +693,11 @@ enum Result {
     FoundSpace
 };
 
-Result hasPreviousSibling(KoXmlNode node)
+Result hasPreviousSibling(QDomNode node)
 {
     while (!node.isNull()) {
         if (node.isElement()) {
-            KoXmlElement element = node.toElement();
+            QDomElement element = node.toElement();
             if (element.tagName() == "text") break;
         }
 
@@ -687,7 +710,7 @@ Result hasPreviousSibling(KoXmlNode node)
             }
 
             if (node.isText()) {
-                KoXmlText textNode = node.toText();
+                QDomText textNode = node.toText();
                 const QString text = cleanUpString(textNode.data());
 
                 if (!text.isEmpty()) {
@@ -709,7 +732,7 @@ Result hasPreviousSibling(KoXmlNode node)
     return FoundNothing;
 }
 
-Result hasNextSibling(KoXmlNode node)
+Result hasNextSibling(QDomNode node)
 {
     while (!node.isNull()) {
         while (!node.nextSibling().isNull()) {
@@ -720,7 +743,7 @@ Result hasNextSibling(KoXmlNode node)
             }
 
             if (node.isText()) {
-                KoXmlText textNode = node.toText();
+                QDomText textNode = node.toText();
                 const QString text = cleanUpString(textNode.data());
 
                 // if we are the trailing whitespace, we should report that
@@ -741,14 +764,12 @@ Result hasNextSibling(KoXmlNode node)
 }
 }
 
-bool KoSvgTextChunkShape::loadSvgTextNode(const KoXmlText &text, SvgLoadingContext &context)
+bool KoSvgTextChunkShape::loadSvgTextNode(const QDomText &text, SvgLoadingContext &context)
 {
-    Q_D(KoSvgTextChunkShape);
-
     SvgGraphicsContext *gc = context.currentGC();
     KIS_SAFE_ASSERT_RECOVER_RETURN_VALUE(gc, false);
 
-    d->loadContextBasedProperties(gc);
+    s->loadContextBasedProperties(gc);
 
     QString data = cleanUpString(text.data());
 
@@ -769,21 +790,18 @@ bool KoSvgTextChunkShape::loadSvgTextNode(const KoXmlText &text, SvgLoadingConte
 
     //ENTER_FUNCTION() << text.data() << "-->" << data;
 
-    d->text = data;
+    s->text = data;
 
     return !data.isEmpty();
 }
 
 void KoSvgTextChunkShape::normalizeCharTransformations()
 {
-    Q_D(KoSvgTextChunkShape);
-    d->applyParentCharTransformations(d->localTransformations);
+    applyParentCharTransformations(s->localTransformations);
 }
 
 void KoSvgTextChunkShape::simplifyFillStrokeInheritance()
 {
-    Q_D(KoSvgTextChunkShape);
-
     if (!isRootTextNode()) {
         KoShape *parentShape = parent();
         KIS_SAFE_ASSERT_RECOVER_RETURN(parentShape);
@@ -823,9 +841,7 @@ void KoSvgTextChunkShape::simplifyFillStrokeInheritance()
 
 KoSvgTextProperties KoSvgTextChunkShape::textProperties() const
 {
-    Q_D(const KoSvgTextChunkShape);
-
-    KoSvgTextProperties properties = d->properties;
+    KoSvgTextProperties properties = s->properties;
     properties.setProperty(KoSvgTextProperties::FillId, QVariant::fromValue(KoSvgText::BackgroundProperty(background())));
     properties.setProperty(KoSvgTextProperties::StrokeId, QVariant::fromValue(KoSvgText::StrokeProperty(stroke())));
 
@@ -834,50 +850,52 @@ KoSvgTextProperties KoSvgTextChunkShape::textProperties() const
 
 bool KoSvgTextChunkShape::isTextNode() const
 {
-    Q_D(const KoSvgTextChunkShape);
     return d->layoutInterface->isTextNode();
 }
 
-KoSvgTextChunkShapeLayoutInterface *KoSvgTextChunkShape::layoutInterface()
+KoSvgTextChunkShapeLayoutInterface *KoSvgTextChunkShape::layoutInterface() const
 {
-    Q_D(KoSvgTextChunkShape);
     return d->layoutInterface.data();
 }
 
+bool KoSvgTextChunkShape::isRichTextPreferred() const
+{
+    return isRootTextNode() && s->isRichTextPreferred;
+}
+
+void KoSvgTextChunkShape::setRichTextPreferred(bool value)
+{
+    KIS_SAFE_ASSERT_RECOVER_RETURN(isRootTextNode());
+
+    s->isRichTextPreferred = value;
+}
+
 bool KoSvgTextChunkShape::isRootTextNode() const
 {
     return false;
 }
 
 /**************************************************************************************************/
-/*                 KoSvgTextChunkShapePrivate                                                     */
+/*                 KoSvgTextChunkShape::Private                                                     */
 /**************************************************************************************************/
 
-#include "SimpleShapeContainerModel.h"
-
-KoSvgTextChunkShapePrivate::KoSvgTextChunkShapePrivate(KoSvgTextChunkShape *_q)
-    : KoShapeContainerPrivate(_q)
+KoSvgTextChunkShape::SharedData::SharedData()
+    : QSharedData()
 {
 }
 
-KoSvgTextChunkShapePrivate::KoSvgTextChunkShapePrivate(const KoSvgTextChunkShapePrivate &rhs, KoSvgTextChunkShape *q)
-    : KoShapeContainerPrivate(rhs, q),
-      properties(rhs.properties),
-      font(rhs.font),
-      fontFamiliesList(rhs.fontFamiliesList),
-      localTransformations(rhs.localTransformations),
-      textLength(rhs.textLength),
-      lengthAdjust(rhs.lengthAdjust),
-      text(rhs.text)
+KoSvgTextChunkShape::SharedData::SharedData(const SharedData &rhs)
+    : QSharedData()
+    , properties(rhs.properties)
+    , localTransformations(rhs.localTransformations)
+    , textLength(rhs.textLength)
+    , lengthAdjust(rhs.lengthAdjust)
+    , text(rhs.text)
+    , isRichTextPreferred(rhs.isRichTextPreferred)
 {
-    if (rhs.model) {
-        SimpleShapeContainerModel *otherModel = dynamic_cast<SimpleShapeContainerModel*>(rhs.model);
-        KIS_ASSERT_RECOVER_RETURN(otherModel);
-        model = new SimpleShapeContainerModel(*otherModel);
-    }
 }
 
-KoSvgTextChunkShapePrivate::~KoSvgTextChunkShapePrivate()
+KoSvgTextChunkShape::SharedData::~SharedData()
 {
 }
 
@@ -885,12 +903,38 @@ KoSvgTextChunkShapePrivate::~KoSvgTextChunkShapePrivate()
 #include <KoColorBackground.h>
 #include <KoShapeStroke.h>
 
-KoSvgText::KoSvgCharChunkFormat KoSvgTextChunkShapePrivate::fetchCharFormat() const
+KoSvgText::KoSvgCharChunkFormat KoSvgTextChunkShape::fetchCharFormat() const
 {
-    Q_Q(const KoSvgTextChunkShape);
-
     KoSvgText::KoSvgCharChunkFormat format;
 
+    const KoSvgTextProperties properties = adjustPropertiesForFontSizeWorkaround(s->properties);
+
+    QFont font(properties.generateFont());
+
+    /**
+     * HACK ALERT: Qt has a bug. When requesting font from the font
+     * database (in QFontDatabase::load()), Qt scales its size by the
+     * current primary display DPI. The only official way to disable
+     * that is to assign QTextDocument to QTextLayout, which is not
+     * something we would like to do. So we do the hackish way, we
+     * just prescale the font with inverted value.
+     *
+     * This hack changes only the rendering process without touching
+     * the way how the text is saved into .kra or .svg. That is nice,
+     * but it means it also affects how old files are renderred. To
+     * let the user open older files we provide a preference option
+     * to enable this scaling again.
+     *
+     * NOTE:  the hack is not needed for pixel-measured fonts, they
+     *        seem to render correctly. Pity we don't use them in
+     *        our SVG code (partially because they don't allow
+     *        fractional-sized fonts).
+     */
+    if (font.pointSizeF() > 0) {
+        qreal adjustedFontSize = 72.0 / qt_defaultDpi() * font.pointSizeF();
+        font.setPointSizeF(adjustedFontSize);
+    }
+
     format.setFont(font);
     format.setTextAnchor(KoSvgText::TextAnchor(properties.propertyOrDefault(KoSvgTextProperties::TextAnchorId).toInt()));
 
@@ -924,7 +968,9 @@ KoSvgText::KoSvgCharChunkFormat KoSvgTextChunkShapePrivate::fetchCharFormat() co
     }
 
     KoSvgText::AutoValue kerning = properties.propertyOrDefault(KoSvgTextProperties::KerningId).value<KoSvgText::AutoValue>();
-    if (!kerning.isAuto) {
+    if (kerning.isAuto) {
+        format.setFontKerning(true);
+    } else {
         format.setFontKerning(false);
         format.setFontLetterSpacingType(QFont::AbsoluteSpacing);
         format.setFontLetterSpacing(format.fontLetterSpacing() + kerning.customValue);
@@ -933,9 +979,10 @@ KoSvgText::KoSvgCharChunkFormat KoSvgTextChunkShapePrivate::fetchCharFormat() co
 
     QBrush textBrush = Qt::NoBrush;
 
-    if (q->background()) {
-        KoColorBackground *colorBackground = dynamic_cast<KoColorBackground*>(q->background().data());
-        KIS_SAFE_ASSERT_RECOVER (colorBackground) {
+    if (background()) {
+        KoColorBackground *colorBackground = dynamic_cast<KoColorBackground*>(background().data());
+        if (!colorBackground) {
+            qWarning() << "TODO: support gradient and pattern backgrounds for text";
             textBrush = Qt::red;
         }
 
@@ -948,8 +995,8 @@ KoSvgText::KoSvgCharChunkFormat KoSvgTextChunkShapePrivate::fetchCharFormat() co
 
     QPen textPen = Qt::NoPen;
 
-    if (q->stroke()) {
-        KoShapeStroke *stroke = dynamic_cast<KoShapeStroke*>(q->stroke().data());
+    if (stroke()) {
+        KoShapeStroke *stroke = dynamic_cast<KoShapeStroke*>(this->stroke().data());
         if (stroke) {
             textPen = stroke->resultLinePen();
         }
@@ -958,19 +1005,17 @@ KoSvgText::KoSvgCharChunkFormat KoSvgTextChunkShapePrivate::fetchCharFormat() co
     format.setTextOutline(textPen);
 
     // TODO: avoid const_cast somehow...
-    format.setAssociatedShape(const_cast<KoSvgTextChunkShape*>(q));
+    format.setAssociatedShape(const_cast<KoSvgTextChunkShape*>(this));
 
     return format;
 }
 
-void KoSvgTextChunkShapePrivate::applyParentCharTransformations(const QVector<KoSvgText::CharTransformation> transformations)
+void KoSvgTextChunkShape::applyParentCharTransformations(const QVector<KoSvgText::CharTransformation> transformations)
 {
-    Q_Q(KoSvgTextChunkShape);
-
-    if (q->shapeCount()) {
+    if (shapeCount()) {
         int numCharsPassed = 0;
 
-        Q_FOREACH (KoShape *shape, q->shapes()) {
+        Q_FOREACH (KoShape *shape, shapes()) {
             KoSvgTextChunkShape *chunkShape = dynamic_cast<KoSvgTextChunkShape*>(shape);
             KIS_SAFE_ASSERT_RECOVER_RETURN(chunkShape);
 
@@ -978,19 +1023,19 @@ void KoSvgTextChunkShapePrivate::applyParentCharTransformations(const QVector<Ko
             QVector<KoSvgText::CharTransformation> t = transformations.mid(numCharsPassed, numCharsInSubtree);
             if (t.isEmpty()) break;
 
-            chunkShape->d_func()->applyParentCharTransformations(t);
+            chunkShape->applyParentCharTransformations(t);
             numCharsPassed += numCharsInSubtree;
 
             if (numCharsPassed >= transformations.size()) break;
         }
     } else {
-        for (int i = 0; i < qMin(transformations.size(), text.size()); i++) {
-            KIS_SAFE_ASSERT_RECOVER_RETURN(localTransformations.size() >= i);
+        for (int i = 0; i < qMin(transformations.size(), s->text.size()); i++) {
+            KIS_SAFE_ASSERT_RECOVER_RETURN(s->localTransformations.size() >= i);
 
-            if (localTransformations.size() == i) {
-                localTransformations.append(transformations[i]);
+            if (s->localTransformations.size() == i) {
+                s->localTransformations.append(transformations[i]);
             } else {
-                localTransformations[i].mergeInParentTransformation(transformations[i]);
+                s->localTransformations[i].mergeInParentTransformation(transformations[i]);
             }
         }
     }
